; LZSS decoder for SNES
; /Mic, 2009


; The base address should be an even multiple of 4KB
.DEFINE LZSS_RAM_BASE 	  $7f0000	
.DEFINE LZSS_TEXTBUF 	  LZSS_RAM_BASE

.DEFINE LZSS_ZP_BASE 	  $80
.DEFINE LZSS_DATA_PTR     LZSS_ZP_BASE
.DEFINE LZSS_FLAGS 	  LZSS_DATA_PTR+3
.DEFINE LZSS_STRLEN 	  LZSS_FLAGS+1
.DEFINE LZSS_TEMP 	  LZSS_STRLEN+1
.DEFINE LZSS_STRPTR 	  LZSS_TEMP+2
.DEFINE LZSS_PTR 	  LZSS_STRPTR+2
.DEFINE LZSS_TEXTBUF_OFFS LZSS_PTR+3
.DEFINE LZSS_LOBYTE	  LZSS_TEXTBUF_OFFS+2
.DEFINE LZSS_FLIPFLOP	  LZSS_LOBYTE+1
.DEFINE LZSS_CNT2 	  LZSS_FLIPFLOP+1

.DEFINE LZSS_CNT 	  LZSS_RAM_BASE+LZSS_DICTIONARY_SIZE
.DEFINE LZSS_BITPLANES    LZSS_CNT2+1
.DEFINE LZSS_PLANECNT 	  LZSS_CNT2+2
.DEFINE LZSS_TEXTBUF_P 	  (LZSS_DICTIONARY_SIZE-LZSS_MAX_LEN)


; Function lzss_decode_vram
;
; Decompress LZSS packed data to VRAM
;
; In:
;  X = Pointer to compressed data (16-bit)
;  Y = Size of compressed data (16-bit)
;  A = VRAM base address (16-bit)
;  DBR = compressed data bank
;
; Out:
;  -
;
; Example:
;
;  sep #$20
;  lda #:packed_patterns
;  phb  
;  pha
;  plb
;  rep #$30
;  lda #$4000
;  ldx #packed_patterns
;  ldy #(packed_patterns_end-packed_patterns)
;  jsr lzss_decode_vram
;  plb
;
lzss_decode_vram:
	php
	pha

	rep	#$30
	
	stz	LZSS_FLIPFLOP
	
	stx	LZSS_DATA_PTR

	ldx	#0
	stx	LZSS_PTR
	ldx	#LZSS_TEXTBUF_P
	stx	LZSS_TEXTBUF_OFFS
	
	; Set the VRAM address
	sta.l 	$002116

	; Increment the VRAM address by one word after each write to $2119	
	sep	#$20
	lda	#$80
	sta.l	$002115

	phb
	pla
	sta	LZSS_DATA_PTR+2
	
	lda	#$7f
	sta	LZSS_PTR+2
	
	; Set all entries in textbuf to 0	
	ldx	#0
	lda	#0
_ldv_init_textbuf:
	sta.l	LZSS_TEXTBUF,x
	inx
	cpx	#LZSS_TEXTBUF_P
	bne	_ldv_init_textbuf

	tyx				; X = size of compressed data
	ldy	#0			; Y = offset into compressed data
_ldv_loop:
	cpx	#0
	bne	+
	jmp	_ldv_done
+:
	lda	(LZSS_DATA_PTR),y	; Read the flags byte
	iny
	dex
	sta	LZSS_FLAGS
	lda	#8
	sta	LZSS_CNT2		; There are 8 flags in the flags byte
_ldv_check_flags:	
	lsr	LZSS_FLAGS
	bcc	_ldv_flag_clear		; Branch if the lowest bit was a zero
	cpx	#0
	bne	+
	jmp	_ldv_done
+:
	lda	#1
	sta	LZSS_STRLEN		; Make sure that only one byte is output
	lda	(LZSS_DATA_PTR),y
	iny
	dex
	phx
	phy
	ldy	LZSS_TEXTBUF_OFFS
	jmp	_ldv_output_byte
_ldv_flag_clear:
	cpx	#0
	beq	_ldv_done
	lda	(LZSS_DATA_PTR),y	; A = idx
	iny
	sta	LZSS_STRPTR
	lda	(LZSS_DATA_PTR),y	; A = j
	iny
	dex
	dex
	sta	LZSS_TEMP+1	
	and	#LZSS_LEN_MASK
	clc
	adc	#LZSS_THRESHOLD+1
	sta	LZSS_STRLEN		; STRLEN = (j & 0x0F) + THRESHOLD + 1
	lda	LZSS_TEMP+1
	.REPT LZSS_LEN_BITS
	lsr	a
	.ENDR
	sta	LZSS_STRPTR+1		; STRPTR = ((j & 0xF0) << 4) | idx
	phx
	phy
	ldx	LZSS_STRPTR
	ldy	LZSS_TEXTBUF_OFFS
_ldv_copy_string:
	lda.l	LZSS_TEXTBUF,x		; Read from TEXTBUF
	pha
	inx
	rep	#$20
	txa
	and	#(LZSS_DICTIONARY_SIZE-1)
	tax
	sep	#$20
	pla
_ldv_output_byte:
	sta	[LZSS_PTR],y		; Write to TEXTBUF
	lda	LZSS_FLIPFLOP
	cmp	#0
	beq	_ldv_write_lo
	lda	[LZSS_PTR],y		
	sta.l	$002119
	jmp	_ldv_wrap_adr
_ldv_write_lo:
	lda	[LZSS_PTR],y
	sta.l	$002118
_ldv_wrap_adr:
	lda	LZSS_FLIPFLOP
	eor	#1
	sta	LZSS_FLIPFLOP
	iny
	rep	#$20
	tya
	and	#(LZSS_DICTIONARY_SIZE-1)
	tay
	sep	#$20
_ldv_next_string_pos:
	dec	LZSS_STRLEN
	bne	_ldv_copy_string
	sty	LZSS_TEXTBUF_OFFS
	ply
	plx
_ldv_next_flag:
	dec	LZSS_CNT2
	bne	+
	jmp	_ldv_loop
+:
	jmp	_ldv_check_flags
_ldv_done:
	rep	#$20
	pla
	plp
	rts
	
	
